import inspect
from collections.abc import Callable, Sequence
from typing import Any, TypeAlias, TypeVar, get_type_hints, overload

from dishka import Provider, Scope

from waku.di._activation import Activator, ConditionalProvider

__all__ = [
    'ProviderSpec',
    'contextual',
    'many',
    'object_',
    'provider',
    'scoped',
    'singleton',
    'transient',
]

ProviderSpec: TypeAlias = Provider | ConditionalProvider

_T = TypeVar('_T')


def provider(
    source: Callable[..., Any] | type[Any],
    *,
    scope: Scope = Scope.REQUEST,
    provided_type: Any | None = None,
    cache: bool = True,
) -> Provider:
    """Create a Dishka provider for a callable or type.

    Args:
        source: Callable or type to provide as a dependency.
        scope: Scope of the dependency (default: Scope.REQUEST).
        provided_type: Explicit type to provide (default: inferred).
        cache: Whether to cache the instance in the scope.

    Returns:
        Provider: Configured provider instance.
    """
    provider_ = Provider(scope=scope)
    provider_.provide(source, provides=provided_type, cache=cache)
    return provider_


@overload
def singleton(source: type[_T] | Callable[..., _T], /) -> Provider: ...


@overload
def singleton(source: type[_T] | Callable[..., _T], /, *, when: Activator) -> ConditionalProvider: ...


@overload
def singleton(interface: Any, implementation: type[_T] | Callable[..., _T], /) -> Provider: ...


@overload
def singleton(
    interface: Any, implementation: type[_T] | Callable[..., _T], /, *, when: Activator
) -> ConditionalProvider: ...


def singleton(
    interface_or_source: type[Any] | Callable[..., Any],
    implementation: type[Any] | Callable[..., Any] | None = None,
    /,
    *,
    when: Activator | None = None,
) -> ProviderSpec:
    """Create a singleton provider (lifetime: app).

    Args:
        interface_or_source: Interface type or source if no separate implementation.
        implementation: Implementation type if interface is provided.
        when: Optional predicate to conditionally activate the provider.

    Returns:
        Provider or ConditionalProvider if `when` is specified.
    """
    if implementation is not None:
        provided_type = interface_or_source
        base = provider(implementation, scope=Scope.APP, provided_type=provided_type)
    else:
        provided_type = _get_provided_type(interface_or_source)
        base = provider(interface_or_source, scope=Scope.APP)

    if when is None:
        return base
    return ConditionalProvider(provider=base, when=when, provided_type=provided_type)


@overload
def scoped(source: type[_T] | Callable[..., _T], /) -> Provider: ...


@overload
def scoped(source: type[_T] | Callable[..., _T], /, *, when: Activator) -> ConditionalProvider: ...


@overload
def scoped(interface: Any, implementation: type[_T] | Callable[..., _T], /) -> Provider: ...


@overload
def scoped(
    interface: Any, implementation: type[_T] | Callable[..., _T], /, *, when: Activator
) -> ConditionalProvider: ...


def scoped(
    interface_or_source: type[Any] | Callable[..., Any],
    implementation: type[Any] | Callable[..., Any] | None = None,
    /,
    *,
    when: Activator | None = None,
) -> ProviderSpec:
    """Create a scoped provider (lifetime: request).

    Args:
        interface_or_source: Interface type or source if no separate implementation.
        implementation: Implementation type if interface is provided.
        when: Optional predicate to conditionally activate the provider.

    Returns:
        Provider or ConditionalProvider if `when` is specified.
    """
    if implementation is not None:
        provided_type = interface_or_source
        base = provider(implementation, scope=Scope.REQUEST, provided_type=provided_type)
    else:
        provided_type = _get_provided_type(interface_or_source)
        base = provider(interface_or_source, scope=Scope.REQUEST)

    if when is None:
        return base
    return ConditionalProvider(provider=base, when=when, provided_type=provided_type)


@overload
def transient(source: type[_T] | Callable[..., _T], /) -> Provider: ...


@overload
def transient(source: type[_T] | Callable[..., _T], /, *, when: Activator) -> ConditionalProvider: ...


@overload
def transient(interface: Any, implementation: type[_T] | Callable[..., _T], /) -> Provider: ...


@overload
def transient(
    interface: Any, implementation: type[_T] | Callable[..., _T], /, *, when: Activator
) -> ConditionalProvider: ...


def transient(
    interface_or_source: type[Any] | Callable[..., Any],
    implementation: type[Any] | Callable[..., Any] | None = None,
    /,
    *,
    when: Activator | None = None,
) -> ProviderSpec:
    """Create a transient provider (new instance per injection).

    Args:
        interface_or_source: Interface type or source if no separate implementation.
        implementation: Implementation type if interface is provided.
        when: Optional predicate to conditionally activate the provider.

    Returns:
        Provider or ConditionalProvider if `when` is specified.
    """
    if implementation is not None:
        provided_type = interface_or_source
        base = provider(implementation, scope=Scope.REQUEST, provided_type=provided_type, cache=False)
    else:
        provided_type = _get_provided_type(interface_or_source)
        base = provider(interface_or_source, scope=Scope.REQUEST, cache=False)

    if when is None:
        return base
    return ConditionalProvider(provider=base, when=when, provided_type=provided_type)


@overload
def object_(obj: Any, *, provided_type: Any | None = None) -> Provider: ...


@overload
def object_(obj: Any, *, provided_type: Any | None = None, when: Activator) -> ConditionalProvider: ...


def object_(
    obj: Any,
    *,
    provided_type: Any | None = None,
    when: Activator | None = None,
) -> ProviderSpec:
    """Provide the exact object passed at creation time as a singleton dependency.

    The provider always returns the same object instance, without instantiation or copying.

    Args:
        obj: The instance to provide as-is.
        provided_type: Explicit type to provide (default: inferred).
        when: Optional predicate to conditionally activate the provider.

    Returns:
        Provider or ConditionalProvider if `when` is specified.
    """
    actual_type = provided_type if provided_type is not None else type(obj)
    base = provider(lambda: obj, scope=Scope.APP, provided_type=actual_type, cache=True)

    if when is None:
        return base
    return ConditionalProvider(provider=base, when=when, provided_type=actual_type)


@overload
def contextual(provided_type: Any, *, scope: Scope = Scope.REQUEST) -> Provider: ...


@overload
def contextual(provided_type: Any, *, scope: Scope = Scope.REQUEST, when: Activator) -> ConditionalProvider: ...


def contextual(
    provided_type: Any,
    *,
    scope: Scope = Scope.REQUEST,
    when: Activator | None = None,
) -> ProviderSpec:
    """Provide a dependency from the current context (e.g., app/request).

    Args:
        provided_type: The type to resolve from context.
        scope: Scope of the context variable (default: Scope.REQUEST).
        when: Optional predicate to conditionally activate the provider.

    Returns:
        Provider or ConditionalProvider if `when` is specified.
    """
    provider_ = Provider()
    provider_.from_context(provided_type, scope=scope)

    if when is None:
        return provider_
    return ConditionalProvider(provider=provider_, when=when, provided_type=provided_type)


def _get_provided_type(impl: Any) -> Any:
    """Extract the type that will be provided by an implementation.

    For classes, returns the class itself.
    For factory functions, returns the return type annotation.

    Raises:
        TypeError: If impl is a factory function without return type annotation.
    """
    if inspect.isclass(impl):
        return impl

    if callable(impl):
        hints = get_type_hints(impl)
        return_type = hints.get('return')
        if return_type is None:
            name = getattr(impl, '__name__', repr(impl))
            msg = f"Factory function '{name}' must have a return type annotation"
            raise TypeError(msg)
        return return_type

    msg = f'Implementation must be a class or callable, got {type(impl).__name__}'
    raise TypeError(msg)


@overload
def many(
    interface: Any,
    *implementations: Any,
    scope: Scope = Scope.REQUEST,
    cache: bool = True,
) -> Provider: ...


@overload
def many(
    interface: Any,
    *implementations: Any,
    scope: Scope = Scope.REQUEST,
    cache: bool = True,
    when: Activator,
) -> ConditionalProvider: ...


def many(
    interface: Any,
    *implementations: Any,
    scope: Scope = Scope.REQUEST,
    cache: bool = True,
    when: Activator | None = None,
) -> ProviderSpec:
    """Register multiple implementations as a collection.

    Args:
        interface: Interface type for the collection.
        *implementations: Implementation types or factory functions to include in collection.
        scope: Scope of the collection (default: Scope.REQUEST).
        cache: Whether to cache the resolve results within scope.
        when: Optional predicate to conditionally activate the provider.

    Returns:
        Provider or ConditionalProvider if `when` is specified.

    Raises:
        ValueError: If no implementations are provided.
        TypeError: If a factory function lacks a return type annotation.
    """
    if not implementations:
        msg = 'At least one implementation must be provided'
        raise ValueError(msg)

    provided_types = [_get_provided_type(impl) for impl in implementations]

    provider_ = Provider(scope=scope)
    provider_.provide_all(*implementations, cache=cache)

    provider_.provide(
        lambda: [],  # noqa: PIE807
        provides=list[interface],
        cache=cache,
    )
    provider_.alias(list[interface], provides=Sequence[interface], cache=cache)

    for provided_type in provided_types:

        @provider_.decorate
        def _(many_: list[interface], one: provided_type) -> list[interface]:  # type: ignore[valid-type]
            return [*many_, one]

    if when is None:
        return provider_
    return ConditionalProvider(provider=provider_, when=when, provided_type=Sequence[interface])



from __future__ import annotations

from contextlib import contextmanager
from itertools import chain
from typing import TYPE_CHECKING, Protocol

from dishka import STRICT_VALIDATION, make_async_container

from waku.di import DEFAULT_COMPONENT, AsyncContainer, BaseProvider

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dishka.dependency_source import Factory
    from dishka.registry import Registry


__all__ = ['override']


class _HasOverride(Protocol):
    override: bool


@contextmanager
def override(container: AsyncContainer, *providers: BaseProvider) -> Iterator[None]:
    """Temporarily override providers in an AsyncContainer for testing.

    Args:
        container: The container whose providers will be overridden.
        *providers: Providers to override in the container.

    Yields:
        None: Context in which the container uses the overridden providers.

    Example:
        ```python
        from waku import WakuFactory, module
        from waku.di import Scope, singleton
        from waku.testing import override


        class Service: ...


        class ServiceOverride(Service): ...


        with override(application.container, singleton(ServiceOverride, provided_type=Service)):
            service = await application.container.get(Service)
            assert isinstance(service, ServiceOverride)
        ```
    """
    for provider in providers:
        for factory in chain[_HasOverride](provider.factories, provider.aliases):
            factory.override = True

    new_container = make_async_container(
        _container_provider(container),
        *providers,
        context=container._context,  # noqa: SLF001
        start_scope=container.scope,
        validation_settings=STRICT_VALIDATION,
    )

    _swap(container, new_container)
    yield
    _swap(new_container, container)


def _container_provider(container: AsyncContainer) -> BaseProvider:
    container_provider = BaseProvider(component=DEFAULT_COMPONENT)
    container_provider.factories.extend(_extract_factories(container.registry))
    for registry in container.child_registries:
        container_provider.factories.extend(_extract_factories(registry))
    return container_provider


def _extract_factories(registry: Registry) -> list[Factory]:
    return [factory for dep_key, factory in registry.factories.items() if dep_key.type_hint is not AsyncContainer]


def _swap(c1: AsyncContainer, c2: AsyncContainer) -> None:
    for attr in type(c1).__slots__:
        tmp = getattr(c1, attr)
        setattr(c1, attr, getattr(c2, attr))
        setattr(c2, attr, tmp)



from __future__ import annotations

import functools
import typing

from typing_extensions import get_original_bases

from waku.cqrs.contracts import ResponseT

if typing.TYPE_CHECKING:
    from waku.cqrs import Request

__all__ = ['get_request_response_type']


@functools.cache
def get_request_response_type(request_type: type[Request[ResponseT]]) -> type[ResponseT]:
    return typing.cast(type[ResponseT], typing.get_args(get_original_bases(request_type)[0])[0])



from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Generic

from waku.cqrs.contracts.request import RequestT, ResponseT

if TYPE_CHECKING:
    from collections.abc import Sequence

    from waku.cqrs.contracts.pipeline import IPipelineBehavior, NextHandlerType


class PipelineBehaviorWrapper(Generic[RequestT, ResponseT]):
    """Composes pipeline behaviors into a processing chain."""

    def __init__(self, behaviors: Sequence[IPipelineBehavior[RequestT, ResponseT]]) -> None:
        """Initialize the pipeline behavior chain.

        Args:
            behaviors: Sequence of pipeline behaviors to execute in order
        """
        self._behaviors = list(behaviors)  # Convert to list immediately

    def wrap(self, handle: NextHandlerType[RequestT, ResponseT]) -> NextHandlerType[RequestT, ResponseT]:
        """Create a pipeline that wraps the handler function with behaviors.

        Pipeline behaviors are executed in the order they are provided.

        Args:
            handle: The handler function to wrap with behaviors

        Returns:
            A function that executes the entire pipeline
        """
        for behavior in reversed(self._behaviors):
            handle = functools.partial(behavior.handle, next_handler=handle)

        return handle



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, NamedTuple, Protocol, TypeAlias

if TYPE_CHECKING:
    from dishka import Provider

    from waku import DynamicModule
    from waku.di._providers import ProviderSpec
    from waku.modules import ModuleType

__all__ = [
    'ActivationBuilder',
    'ActivationContext',
    'Activator',
    'ConditionalProvider',
    'Has',
    'IProviderFilter',
    'ProviderFilter',
]


class ActivationBuilder(Protocol):
    @abstractmethod
    def has_active(self, type_: Any) -> bool:
        raise NotImplementedError


class ActivationContext(NamedTuple):
    """Context passed to activators for provider activation decisions."""

    container_context: dict[Any, Any] | None
    module_type: ModuleType | DynamicModule
    provided_type: Any
    builder: ActivationBuilder


Activator: TypeAlias = Callable[[ActivationContext], bool]


@dataclass(frozen=True, slots=True)
class Has:
    """Activator that checks if a provider for a type is registered."""

    type_: Any

    def __call__(self, ctx: ActivationContext) -> bool:
        return ctx.builder.has_active(self.type_)


@dataclass(frozen=True, slots=True)
class ConditionalProvider:
    """Provider with activation condition."""

    provider: Provider
    when: Activator
    provided_type: Any


class IProviderFilter(Protocol):
    """Strategy for filtering providers based on activation context."""

    def filter(
        self,
        providers: list[ProviderSpec],
        context: dict[Any, Any] | None,
        module_type: ModuleType | DynamicModule,
        builder: ActivationBuilder,
    ) -> list[Provider]: ...


OnSkipCallback: TypeAlias = Callable[[ConditionalProvider, ActivationContext], None]


@dataclass(slots=True)
class ProviderFilter:
    """Default provider filter implementation."""

    on_skip: OnSkipCallback | None = field(default=None)

    def filter(
        self,
        providers: list[ProviderSpec],
        context: dict[Any, Any] | None,
        module_type: ModuleType | DynamicModule,
        builder: ActivationBuilder,
    ) -> list[Provider]:
        result: list[Provider] = []

        for spec in providers:
            if isinstance(spec, ConditionalProvider):
                ctx = ActivationContext(
                    container_context=context,
                    module_type=module_type,
                    provided_type=spec.provided_type,
                    builder=builder,
                )
                if spec.when(ctx):
                    result.append(spec.provider)
                elif self.on_skip:
                    self.on_skip(spec, ctx)
            else:
                result.append(spec)

        return result



from __future__ import annotations

from collections import OrderedDict, defaultdict
from typing import TYPE_CHECKING, Any, Final, TypeAlias
from uuid import UUID

from waku.di import ConditionalProvider, IProviderFilter, ProviderFilter
from waku.modules import Module, ModuleCompiler, ModuleMetadata, ModuleRegistry, ModuleType

if TYPE_CHECKING:
    from waku import DynamicModule
    from waku.di import BaseProvider


__all__ = [
    'AdjacencyMatrix',
    'ModuleRegistryBuilder',
]


AdjacencyMatrix: TypeAlias = dict[UUID, OrderedDict[UUID, str]]


class _ActivationBuilder:
    """Build-time activation builder for checking registered types."""

    def __init__(self) -> None:
        self._registered_types: set[Any] = set()

    def register(self, type_: Any) -> None:
        self._registered_types.add(type_)

    def has_active(self, type_: Any) -> bool:
        return type_ in self._registered_types


class ModuleRegistryBuilder:
    def __init__(
        self,
        root_module_type: ModuleType,
        compiler: ModuleCompiler | None = None,
        context: dict[Any, Any] | None = None,
        provider_filter: IProviderFilter | None = None,
    ) -> None:
        self._compiler: Final = compiler or ModuleCompiler()
        self._root_module_type: Final = root_module_type
        self._context: Final = context
        self._provider_filter: Final[IProviderFilter] = provider_filter or ProviderFilter()
        self._modules: dict[UUID, Module] = {}
        self._providers: list[BaseProvider] = []

        self._metadata_cache: dict[ModuleType | DynamicModule, tuple[ModuleType, ModuleMetadata]] = {}
        self._builder: Final = _ActivationBuilder()

    def build(self) -> ModuleRegistry:
        modules, adjacency = self._collect_modules()
        self._build_type_registry(modules)
        root_module = self._register_modules(modules)
        return self._build_registry(root_module, adjacency)

    def _collect_modules(self) -> tuple[list[tuple[ModuleType, ModuleMetadata]], AdjacencyMatrix]:
        """Collect modules in post-order DFS."""
        visited: set[UUID] = set()
        post_order: list[tuple[ModuleType, ModuleMetadata]] = []
        adjacency: AdjacencyMatrix = defaultdict(OrderedDict)
        self._collect_modules_recursive(self._root_module_type, visited, post_order, adjacency)
        return post_order, adjacency

    def _collect_modules_recursive(
        self,
        current_type: ModuleType | DynamicModule,
        visited: set[UUID],
        post_order: list[tuple[ModuleType, ModuleMetadata]],
        adjacency: AdjacencyMatrix,
    ) -> None:
        type_, metadata = self._get_metadata(current_type)
        if metadata.id in visited:
            return

        adjacency[metadata.id][metadata.id] = type_.__name__

        for imported in metadata.imports:
            imported_type, imported_metadata = self._get_metadata(imported)
            adjacency[metadata.id][imported_metadata.id] = imported_type.__name__
            if imported_metadata.id not in visited:
                self._collect_modules_recursive(imported, visited, post_order, adjacency)

        post_order.append((type_, metadata))
        visited.add(metadata.id)

    def _build_type_registry(self, modules: list[tuple[ModuleType, ModuleMetadata]]) -> None:
        """Build registry of all provided types before filtering."""
        for _, metadata in modules:
            for spec in metadata.providers:
                if isinstance(spec, ConditionalProvider):
                    self._builder.register(spec.provided_type)
                else:
                    for factory in spec.factories:
                        self._builder.register(factory.provides.type_hint)

    def _register_modules(self, post_order: list[tuple[ModuleType, ModuleMetadata]]) -> Module:
        for type_, metadata in post_order:
            if metadata.id in self._modules:
                continue

            if type_ is self._root_module_type:
                metadata.is_global = True

            module = Module(type_, metadata)

            self._modules[module.id] = module
            self._providers.append(
                module.create_provider(
                    context=self._context,
                    builder=self._builder,
                    provider_filter=self._provider_filter,
                )
            )

        _, root_metadata = self._get_metadata(self._root_module_type)
        return self._modules[root_metadata.id]

    def _get_metadata(self, module_type: ModuleType | DynamicModule) -> tuple[ModuleType, ModuleMetadata]:
        """Get metadata with caching to avoid repeated extractions."""
        if module_type not in self._metadata_cache:
            self._metadata_cache[module_type] = self._compiler.extract_metadata(module_type)
        return self._metadata_cache[module_type]

    def _build_registry(self, root_module: Module, adjacency: AdjacencyMatrix) -> ModuleRegistry:
        return ModuleRegistry(
            compiler=self._compiler,
            modules=self._modules,
            providers=self._providers,
            root_module=root_module,
            adjacency=adjacency,
        )



from __future__ import annotations

import uuid
from dataclasses import dataclass, field
from typing import TypeVar

EventT = TypeVar('EventT', bound='Event', contravariant=True)  # noqa: PLC0105


@dataclass(frozen=True, kw_only=True)
class Event:
    """Base class for events."""

    event_id: uuid.UUID = field(default_factory=uuid.uuid4)



from __future__ import annotations

from collections import defaultdict
from collections.abc import MutableMapping
from typing import Any, Self

from waku.cqrs.contracts.pipeline import IPipelineBehavior
from waku.cqrs.contracts.request import RequestT, ResponseT
from waku.cqrs.exceptions import PipelineBehaviorAlreadyRegistered

PipelineBehaviorMapRegistry = MutableMapping[type[RequestT], list[type[IPipelineBehavior[RequestT, ResponseT]]]]  # ty: ignore[invalid-argument-type]


class PipelineBehaviorMap:
    def __init__(self) -> None:
        self._registry: PipelineBehaviorMapRegistry[Any, Any] = defaultdict(list)

    def bind(
        self,
        request_type: type[RequestT],
        behavior_types: list[type[IPipelineBehavior[RequestT, ResponseT]]],
    ) -> Self:
        for behavior_type in behavior_types:
            if behavior_type in self._registry[request_type]:
                raise PipelineBehaviorAlreadyRegistered(request_type, behavior_type)
            self._registry[request_type].append(behavior_type)
        return self

    def merge(self, other: PipelineBehaviorMap) -> Self:
        for event_type, handlers in other.registry.items():
            self.bind(event_type, handlers)
        return self

    @property
    def registry(self) -> PipelineBehaviorMapRegistry[Any, Any]:
        return self._registry

    def __bool__(self) -> bool:
        return bool(self._registry)



from __future__ import annotations

import uuid
from dataclasses import dataclass, field
from typing import Any, Generic

from typing_extensions import TypeVar

__all__ = [
    'Request',
    'RequestT',
    'Response',
    'ResponseT',
]


RequestT = TypeVar('RequestT', bound='Request[Any]', contravariant=True)  # noqa: PLC0105
ResponseT = TypeVar('ResponseT', bound='Response | None', default=None, covariant=True)  # noqa: PLC0105


@dataclass(frozen=True, kw_only=True)
class Request(Generic[ResponseT]):
    """Base class for request-type objects."""

    request_id: uuid.UUID = field(default_factory=uuid.uuid4)


@dataclass(frozen=True, kw_only=True)
class Response:
    """Base class for response type objects."""



from __future__ import annotations

from typing import TYPE_CHECKING, Final, cast

from waku.di import DEFAULT_COMPONENT, ActivationBuilder, BaseProvider, IProviderFilter

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from typing import Any
    from uuid import UUID

    from dishka import Provider

    from waku.di import ProviderSpec
    from waku.extensions import ModuleExtension
    from waku.modules._metadata import DynamicModule, ModuleMetadata, ModuleType


__all__ = ['Module']


class Module:
    __slots__ = (
        '_provider',
        'exports',
        'extensions',
        'id',
        'imports',
        'is_global',
        'providers',
        'target',
    )

    def __init__(self, module_type: ModuleType, metadata: ModuleMetadata) -> None:
        self.id: Final[UUID] = metadata.id
        self.target: Final[ModuleType] = module_type

        self.providers: Final[Sequence[ProviderSpec]] = metadata.providers
        self.imports: Final[Sequence[ModuleType | DynamicModule]] = metadata.imports
        self.exports: Final[Sequence[type[object] | ModuleType | DynamicModule]] = metadata.exports
        self.extensions: Final[Sequence[ModuleExtension]] = metadata.extensions
        self.is_global: Final[bool] = metadata.is_global

        self._provider: BaseProvider | None = None

    @property
    def name(self) -> str:
        return self.target.__name__

    @property
    def provider(self) -> BaseProvider:
        """Get the aggregated provider for this module.

        This property returns the provider created by create_provider().
        Must be called after create_provider() has been invoked.

        Raises:
            RuntimeError: If create_provider() has not been called yet.
        """
        if self._provider is None:
            msg = f'Module {self.name} provider not yet created. Call create_provider() first.'
            raise RuntimeError(msg)
        return self._provider

    def create_provider(
        self,
        context: dict[Any, Any] | None,
        builder: ActivationBuilder,
        provider_filter: IProviderFilter,
    ) -> BaseProvider:
        """Create aggregated provider with activation filtering applied.

        Args:
            context: Context dict for activation decisions.
            builder: Activation builder for checking if types are registered.
            provider_filter: Filter strategy for conditional provider activation.

        Returns:
            BaseProvider with only active providers aggregated.
        """
        active_providers = provider_filter.filter(
            list(self.providers),
            context=context,
            module_type=self.target,
            builder=builder,
        )

        cls = cast(type[_ModuleProvider], type(f'{self.name}Provider', (_ModuleProvider,), {}))
        self._provider = cls(active_providers)
        return self._provider

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f'Module[{self.name}]'

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        return self.id == other.id if isinstance(other, Module) else False


class _ModuleProvider(BaseProvider):
    """Aggregates factories from filtered providers."""

    def __init__(self, providers: Iterable[Provider]) -> None:
        super().__init__(DEFAULT_COMPONENT)
        for provider in providers:
            self.factories.extend(provider.factories)
            self.aliases.extend(provider.aliases)
            self.decorators.extend(provider.decorators)
            self.context_vars.extend(provider.context_vars)



from waku.cqrs.contracts.event import Event, EventT
from waku.cqrs.contracts.pipeline import IPipelineBehavior, NextHandlerType
from waku.cqrs.contracts.request import Request, RequestT, Response, ResponseT

__all__ = [
    'Event',
    'EventT',
    'IPipelineBehavior',
    'NextHandlerType',
    'Request',
    'RequestT',
    'Response',
    'ResponseT',
]



from collections.abc import Sequence
from dataclasses import dataclass
from itertools import chain
from typing import Any, Self, TypeAlias

from waku.cqrs.contracts.event import EventT
from waku.cqrs.contracts.pipeline import IPipelineBehavior
from waku.cqrs.contracts.request import RequestT
from waku.cqrs.events.handler import EventHandler
from waku.cqrs.events.map import EventMap
from waku.cqrs.events.publish import EventPublisher, SequentialEventPublisher
from waku.cqrs.impl import Mediator
from waku.cqrs.interfaces import IMediator
from waku.cqrs.pipeline.map import PipelineBehaviorMap
from waku.cqrs.requests.handler import RequestHandler
from waku.cqrs.requests.map import RequestMap
from waku.cqrs.utils import get_request_response_type
from waku.di import Provider, WithParents, many, scoped, transient
from waku.extensions import OnModuleConfigure
from waku.modules import DynamicModule, ModuleMetadata, module

__all__ = [
    'MediatorConfig',
    'MediatorExtension',
    'MediatorModule',
]


_HandlerProviders: TypeAlias = tuple[Provider, ...]


@dataclass(frozen=True, slots=True, kw_only=True)
class MediatorConfig:
    """Configuration for the Mediator extension.

    This class defines the configuration options for setting up the cqrs pattern
    implementation in the application.

    Attributes:
        mediator_implementation_type: The concrete implementation class for the cqrs
            interface (IMediator). Defaults to the standard Mediator class.
        event_publisher: The implementation class for publishing events. Defaults to `SequentialEventPublisher`.

        pipeline_behaviors: A sequence of pipeline behavior configurations that will be applied
            to the cqrs pipeline. Behaviors are executed in the order they are defined.
            Defaults to an empty sequence.

    Example:
        ```python
        config = MediatorConfig(
            pipeline_behaviors=[
                LoggingBehavior,
                ValidationBehavior,
            ]
        )
        ```
    """

    mediator_implementation_type: type[IMediator] = Mediator
    event_publisher: type[EventPublisher] = SequentialEventPublisher
    pipeline_behaviors: Sequence[type[IPipelineBehavior[Any, Any]]] = ()


@module()
class MediatorModule:
    @classmethod
    def register(cls, config: MediatorConfig | None = None, /) -> DynamicModule:
        """Application-level module for Mediator setup.

        Args:
            config: Configuration for the Mediator extension.
        """
        config_ = config or MediatorConfig()
        return DynamicModule(
            parent_module=cls,
            providers=[
                *cls._create_mediator_providers(config_),
                *cls._create_pipeline_behavior_providers(config_),
            ],
            is_global=True,
        )

    @staticmethod
    def _create_mediator_providers(config: MediatorConfig) -> _HandlerProviders:
        return (
            scoped(WithParents[IMediator], config.mediator_implementation_type),  # ty: ignore[non-subscriptable]
            scoped(EventPublisher, config.event_publisher),
        )

    @staticmethod
    def _create_pipeline_behavior_providers(config: MediatorConfig) -> _HandlerProviders:
        if not config.pipeline_behaviors:
            return ()
        return (many(IPipelineBehavior[Any, Any], *config.pipeline_behaviors),)


class MediatorExtension(OnModuleConfigure):
    def __init__(self) -> None:
        self._request_map = RequestMap()
        self._event_map = EventMap()
        self._behavior_map = PipelineBehaviorMap()

    def bind_request(
        self,
        request_type: type[RequestT],
        handler_type: type[RequestHandler[RequestT, Any]],
        *,
        behaviors: list[type[IPipelineBehavior[RequestT, Any]]] | None = None,
    ) -> Self:
        self._request_map.bind(request_type, handler_type)
        if behaviors:
            self._behavior_map.bind(request_type, behaviors)
        return self

    def bind_event(
        self,
        event_type: type[EventT],
        handler_types: list[type[EventHandler[EventT]]],
    ) -> Self:
        self._event_map.bind(event_type, handler_types)
        return self

    def on_module_configure(self, metadata: ModuleMetadata) -> None:
        metadata.providers.extend(
            chain(
                self._create_request_handler_providers(),
                self._create_event_handler_providers(),
                self._create_pipeline_behavior_providers(),
            ),
        )

    def _create_request_handler_providers(self) -> _HandlerProviders:
        return tuple(
            transient(
                RequestHandler[request_type, get_request_response_type(request_type)],  # type: ignore[arg-type, valid-type, misc]
                handler_type,
            )
            for request_type, handler_type in self._request_map.registry.items()
        )

    def _create_event_handler_providers(self) -> _HandlerProviders:
        return tuple(
            many(EventHandler[event_type], *handler_types)  # type: ignore[valid-type]
            for event_type, handler_types in self._event_map.registry.items()
        )

    def _create_pipeline_behavior_providers(self) -> _HandlerProviders:
        return tuple(
            many(
                IPipelineBehavior[request_type, get_request_response_type(request_type)],  # type: ignore[arg-type, valid-type, misc]
                *behavior_types,
            )
            for request_type, behavior_types in self._behavior_map.registry.items()
        )



from __future__ import annotations

from collections.abc import MutableMapping
from typing import Any, Self, TypeAlias

from waku.cqrs.contracts.request import RequestT, ResponseT
from waku.cqrs.exceptions import RequestHandlerAlreadyRegistered
from waku.cqrs.requests.handler import RequestHandler

__all__ = [
    'RequestMap',
    'RequestMapRegistry',
]

RequestMapRegistry: TypeAlias = MutableMapping[type[RequestT], type[RequestHandler[RequestT, ResponseT]]]


class RequestMap:
    def __init__(self) -> None:
        self._registry: RequestMapRegistry[Any, Any] = {}

    def bind(
        self,
        request_type: type[RequestT],
        handler_type: type[RequestHandler[RequestT, ResponseT]],
    ) -> Self:
        if request_type in self._registry:
            raise RequestHandlerAlreadyRegistered(request_type, handler_type)
        self._registry[request_type] = handler_type
        return self

    def merge(self, other: RequestMap) -> Self:
        for request_type, handler_type in other.registry.items():
            self.bind(request_type, handler_type)
        return self

    @property
    def registry(self) -> RequestMapRegistry[Any, Any]:
        return self._registry

    def __bool__(self) -> bool:
        return bool(self._registry)



from waku.modules._metadata import DynamicModule, HasModuleMetadata, ModuleCompiler, ModuleMetadata, ModuleType, module
from waku.modules._module import Module
from waku.modules._registry import ModuleRegistry
from waku.modules._registry_builder import ModuleRegistryBuilder

__all__ = [
    'DynamicModule',
    'HasModuleMetadata',
    'Module',
    'ModuleCompiler',
    'ModuleMetadata',
    'ModuleRegistry',
    'ModuleRegistryBuilder',
    'ModuleType',
    'module',
]



from waku.validation._abc import ValidationRule
from waku.validation._errors import ValidationError
from waku.validation._extension import ValidationExtension

__all__ = [
    'ValidationError',
    'ValidationExtension',
    'ValidationRule',
]
